#ifdef __linux
/* These guys are nuts... */
#define _XOPEN_SOURCE 500
#define _BSD_SOURCE
/* 
 * Don't use this unless I have to:
 * #define _GNU_SOURCE
 */
/* 
 * Let me see if I get that straight: 
 * -first enable "some" additional calls
 */
#define _LARGEFILE_SOURCE
/* 
 * -then enable the 64 bits variants of traditional calls
 */
#define _LARGEFILE64_SOURCE
/*
 * and now make the 64 bit variants shadow the old calls
 */
#define _FILE_OFFSET_BITS 64
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <glob.h>
#include <fcntl.h>

#include "my_strptime.h"

typedef char *(*strptime_ft)(const char *, const char *, struct tm *tm);

struct options {
  char t;           /* Subtype of operation */
  char **argv;      /* Copy of commandline arguments */
  int argc;         /*   "                           */
  strptime_ft current_strptime;
};

char *
strcatcat(const int argc, char *const argv[])
{
  char *buf;
  int length = 0;
  int i;
  const char *qs;
  char *ts;

  for (i = 0; i < argc; i++)
    length += strlen(argv[argc-1]) + 1;
  buf = malloc(length + 1);
  if (buf == NULL) {
    perror("malloc");
    exit(2);
  }

  ts = buf;
  for (i = 0; i < argc; i++) {
    for (qs = argv[i]; *qs != '\0'; qs++) {
      *ts++ = *qs;
    }
    *ts++ = ' ';
  }
  *ts++ = '\0';
  return buf;
}

static void usage(void)
#ifdef __GNUC__
     __attribute__ ((noreturn))
#endif
     ;

static void
usage(void)
{
  fprintf(stderr, "Usage:\n"
	  "-t s                    print seconds since 1970 for current time\n"
	  "-t s <datespec>         print seconds since 1970 for <datespec>\n"
	  "   supported formats (only C locale, no national locales!):\n"
	  "      date(1)           Tue May 16 17:10:19 MEST 2000\n"
	  "      apache-log        16/May/2000:17:07:01\n"
	  "      ISO-8601 short    20000516T151152\n"
	  "      Random U.S. crap  02-21-1999 02:31 PM (month first, 12h)\n"
	  "-t d <seconds>          print date(1)-format for 1970+sec \n"
	  "-t a <seconds>          print apache-format for 1970+sec \n"
	  "-t i <seconds>          print ISO-8601 short for 1970+sec \n"
	  "-t u <seconds>          print Unix-like CCYYMMDDhhmm.SS "
	  "for 1970+sec \n"
	  "-f a <file>             file access date in secs from 1970\n"
	  "-f c <file>             file status date in secs from 1970\n"
	  "-f m <file>             file modification date in secs from 1970\n"
	  "-f l <symlink>          print target of symbolic link\n"
	  "-f P <file> <blocksz>   Pad file to blocksize with zeros\n"
	  "-f w <file>             wipeout (overwrite with zeros)\n"
	  "-c                      count from 1\n"
	  "-c <n>                  count from 1 to <n>\n"
	  "-c <n> <m>              count from <n> to <m>\n"
	  "   if <n> has a leading zero, all number will have zeros in front\n"
	  "   so that the length of each number is that of <m> (not <n>)\n"
	  "-c <n> <m> <o>          dito, with step <o> (may be negative)\n"
	  "-r i                    random integer between 0 and 2147483647\n"
	  "-r i <n>                random integer number between 0 and <n>\n"
	  "-r f                    random float between 0.0 and 1.0\n"
	  "-r f <f>                random float between 0.0 and <f>\n"
	  "-r a <arguments ...>    echo arguments in random order\n"
	  "-r a                    as before, but use * (without dirs)\n"
	  );
  exit(1);
}

static int
do_unknown(const struct options *const o)
{
  usage();
}

static char *
Strptime(const struct options *const o, const char *const s
	 , const char *const format,  struct tm *const tm)
{
  char *ret;

  bzero(tm, sizeof(*tm));
  ret = o->current_strptime(s, format, tm);
  if (ret == NULL || mktime(tm) == -1)
    return NULL;
  return ret;
}


static int
do_time(const struct options *const o)
{
  time_t t;
  struct tm tm;
  struct tm *ptm;
  char *buf;
  struct {
    char *fmt;
    int nativeonly;
  } *it, strptimetests[] = {
    /* date(1) */
    {"%+", 0},
    {"%a %Ef %X %Z %Y", 0},
    {"%a %B %e %H:%M:%S %Z %G", 0},
    /* ./shellsupport -ts '20020913 04:59' */
    {"%Y%m%d %H:%M", 0},
    /* '12sept2002 00:00' */
    {"%d%b%Y %H:%M", 0},
#ifdef __GLIBC__
    {"%C", 1},
#endif
    {"%c", 0},
    {"%d/%b/%Y:%T", 0},

    {"%Y%m%dT%H%M%S", 0},
    {"%m-%d-%Y %I:%M:%S %p", 0},
    {"%m-%d-%Y %I:%M %p", 0},
    /* Unix-style [[CC]YY]MMDDhhmm[.SS]] */
    {"%Y%m%d%H%M.%S", 0},
    {"%Y%m%d%H%M", 0},
    {"%y%m%d%H%M.%S", 0},
    {"%y%m%d%H%M", 0},
    /* ./shellsupport -t s 'February 01, 2002 12:28 pm' */
    {"%h %e, %Y %I:%M %p", 0},
    {NULL}
  };
  int success;

  bzero(&tm, sizeof(tm));
  switch (o->t) {
  case 's':
    if (o->argc == 0)
      printf("%.0f\n", (double)time(NULL));
    else {
      buf = strcatcat(o->argc, o->argv);
      success = 0;
      for (it = strptimetests; it->fmt; it++) {
#ifdef HAVE_STRPTIME
	if (!(o->current_strptime == strptime && it->nativeonly)) {
#else
	if (!it->nativeonly) {
#endif
	  if (Strptime(o, buf, it->fmt, &tm)) {
	    success = 1;
	    break;
	  }
	}
      }
      if (success == 0) {
	fprintf(stderr, "Cannot parse time: '%s'\n", buf);
	exit(1);
      }
      free(buf);
      printf("%.0f\n", (double)mktime(&tm));
    }
    break;
  case 'd':
    if (o->argc != 1)
      usage();
    t = atol(o->argv[0]);
    printf("%s", ctime(&t));
    break;
  case 'a':
    if (o->argc != 1)
      usage();
    t = atol(o->argv[0]);
    buf = malloc(8192);
    if (buf == NULL) {
      perror("malloc");
      exit(2);
    }
    ptm = gmtime(&t);
    if (strftime(buf, 8192, "%d/%b/%Y:%T", ptm) == 0) {
      fprintf(stderr, "strftime failed\n");
      exit(2);
    }
    printf("%s\n", buf);
    free(buf);
    break;
  case 'i':
    if (o->argc != 1)
      usage();
    t = atol(o->argv[0]);
    buf = malloc(32);
    if (buf == NULL) {
      perror("malloc");
      exit(2);
    }
    ptm = gmtime(&t);
    if (strftime(buf, 32, "%Y%m%dT%H%M%S", ptm) == 0) {
      fprintf(stderr, "Strftime failed\n");
      exit(2);
    }
    printf("%s\n", buf);
    free(buf);
    break;
  case 'r':
    if (o->argc != 1)
      usage();
    t = atol(o->argv[0]);
    buf = malloc(32);
    if (buf == NULL) {
      perror("malloc");
      exit(2);
    }
    ptm = gmtime(&t);
    if (strftime(buf, 32, "%Y%m%d %H:%M", ptm) == 0) {
      fprintf(stderr, "Strftime failed\n");
      exit(2);
    }
    printf("%s\n", buf);
    free(buf);
    break;
  case 'u':
    if (o->argc != 1)
      usage();
    t = atol(o->argv[0]);
    buf = malloc(32);
    if (buf == NULL) {
      perror("malloc");
      exit(2);
    }
    ptm = gmtime(&t);
    if (strftime(buf, 32, "%Y%m%d%H%M.%S", ptm) == 0) {
      fprintf(stderr, "Strftime failed\n");
      exit(2);
    }
    printf("%s\n", buf);
    free(buf);
    break;
  default:
    fprintf(stderr, "Fell through date output format specs\n");
    usage();
  }
  return 0;
}

static int
do_file(const struct options *const o)
{
  struct stat sb;
  char linktarget[8192];
  long long blocksize = 0;

  if (o->t == 'P') {
    if (o->argc != 2) {
      usage();
    }
    blocksize = atoi(o->argv[1]);
    if (blocksize < 2) {
      fprintf(stderr, "Blocksize for padding must be > 1\n");
      exit(1);
    }
  } else {
    if (o->argc != 1)
      usage();
  }

  if (o->t != 'l' && stat(o->argv[0], &sb) == -1) {
    perror("stat");
    exit(2);
  }

  switch (o->t) {
  case 'a':
    printf("%.0f\n", (double)sb.st_atime);
    break;
  case 'c':
    printf("%.0f\n", (double)sb.st_ctime);
    break;
  case 'm':
    printf("%.0f\n", (double)sb.st_mtime);
    break;
  case 'l':
    if (readlink(o->argv[0], linktarget, sizeof(linktarget)) == -1) {
      perror("readlink");
      exit(2);
    }
    printf("%s\n", linktarget);
    break;
  case 'P': {
    FILE *f;
    int i;
    fprintf(stderr, "Size; %qd -> %qd\n",
	    sb.st_size, 
	    ((sb.st_size / blocksize) + (long long)1) * blocksize
	    );
    f = fopen(o->argv[0], "a");
    if (f == NULL) {
      perror("open");
      exit(2);
    }

    for (i = 0;
	 i < ((sb.st_size / blocksize) + (long long)1) * blocksize -sb.st_size;
	 i++)
      if (fputc('\0', f) == EOF) {
	perror("fputc");
	exit(2);
      }
    if (fclose(f) != 0) {
      perror("fclose");
      exit(2);
    }
    break;
  }
  case 'w': {
    int fd;
    int i;
    char buf[8192];
    bzero(buf, sizeof(buf));

    fd = open(o->argv[0], O_WRONLY);
    if (fd == -1) {
      perror("open");
      exit(2);
    }

    for (i = 0;
	 i < (off_t)sb.st_size / (off_t)sizeof(buf);
	 i++) {
      if (write(fd, buf, sizeof(buf)) == -1) {
	perror("write error - ignored in wipe mode, continuing");
      }
    }
    if (write(fd, buf, (off_t)sb.st_size - 
	      (off_t)sb.st_size / (off_t)sizeof(buf) * (off_t)sizeof(buf)) 
	== -1) {
      perror("write error - ignored in wipe mode, continuing");
    }

    if (close(fd) == -1) {
      perror("close");
      exit(2);
    }
    break;
  }
  default:
    usage();
  }
  return 0;
}

static int
do_count(const struct options *const o)
{
  int i;
  int to;
  char format[256];
  int endless = 0;
  int step = 1;

  strncpy(format, "%d\n", sizeof(format) - 1);
  if (o->argc == 0) {
    i = 1;
    endless = 1;
    to = 1; /* Kill warning */
  } else if (o->argc == 1) {
    i = 1;
    to = atoi(o->argv[0]);
  } else {
    /* Has a 'from' specification */
    if (o->argv[0][0] == '0' && o->argv[0][1] != '\0') {
      /* Leading zero? */
      snprintf(format, sizeof(format) -1, "%%0%dd\n", strlen(o->argv[1]));
    }
    if (o->argc == 2) {
      i = atoi(o->argv[0]);
      to = atoi(o->argv[1]);
    } else if (o->argc == 3) {
      i = atoi(o->argv[0]);
      to = atoi(o->argv[1]);
      step = atoi(o->argv[2]);
    } else {
      usage();
    }
  }
  if (step == 0)
    usage();
  for (;; i += step) {
    if (!endless && ((step > 0 && i > to) || (step < 0 && i < to)))
      break;
    printf(format, i);
  }
  return 0;
}

static int
globerr(const char *const s, const int code)
{
  fprintf(stderr, "glob error at '%s' with '%d'\n", s, code);
  return 1;
}

static void
do_random_echo(const struct options *const o)
{
  const char **target;
  int i;
  int pos;
  int count;
  char **args;
  int is_glob = 0;
  glob_t g;
  struct stat sb;

  if (o->argc == 0) {
    if (glob("*", GLOB_ERR, globerr, &g) != 0) {
      fprintf(stderr, "glob failed ");
      perror(NULL);
      exit(2);
    }
    is_glob = 1;

    count = g.gl_pathc;
    args = g.gl_pathv;
  } else {
    count = o->argc;
    args = o->argv;
  }

  target = calloc(1, sizeof(char *) * count);
  if (target == NULL) {
    perror("malloc");
    exit(2);
  }

  /* Not efficient, but sufficient since argument list length is limited */
  for (i = 0; i < count; i++) {
    for (pos = random() % count;
	 target[pos] != NULL && pos < count; 
	 pos = random() % count) {
    }
    target[pos] = args[i];
  }

  for (i = 0; i < count; i++) {
    if (is_glob) {
      if (stat(target[i], &sb) == -1) {
	fprintf(stderr, "stat '%s' ", target[i]);
	perror(NULL);
	exit(2);
      }
      if (sb.st_mode & S_IFDIR)
	continue;
    }
    printf("%s\n", target[i]);
  }

  free(target);
  if (is_glob)
    globfree(&g);
}

static int
do_random(const struct options *const o)
{
  switch (o->t) {
  case 'i':
    if (o->argc > 1)
      usage();
    if (o->argc == 0)
      printf("%ld\n", 
#ifdef __linux
	     /* Idiots... */
	     (long)
#endif
	     random());
    else
      printf("%ld\n", random() % atol(o->argv[0]));
    break;
  case 'f':
    if (o->argc > 1)
      usage();
    if (o->argc == 0)
      printf("%g\n", (double)random() / (double)RAND_MAX * 1.0);
    else
      printf("%g\n", (double)random() / (double)RAND_MAX * atof(o->argv[0]));
    break;
  case 'a':
    do_random_echo(o);
    break;
  default:
    usage();
  }
  return 0;
}

static void
default_options(struct options *const o)
{
  const char *e;
  bzero(o, sizeof(*o));
  e = getenv("SHELLSUPPORT_STRPTIME");
  if (e && !strcmp(e, "native")) {
#ifdef HAVE_STRPTIME
    o->current_strptime = strptime;
#else
    fprintf(stderr, "Don't have native strptime\n");
    exit(1);
#endif
  } else {
    o->current_strptime = my_strptime;
  }
}

int
main(int argc, char *argv[])
{
  int ch;
  struct options o;
  int (*func)(const struct options *o);
  func = do_unknown;

  default_options(&o);

  srandom(time(NULL) ^ getpid());

  while ((ch = getopt(argc, argv, "cf:r:R:t:")) != -1) {
    switch (ch) {
    case 'c':
      func = do_count;
      break;
    case 'f':
      o.t = optarg[0];
      func = do_file;
      break;
    case 'r':
      o.t = optarg[0];
      func = do_random;
      break;
    case 'R':
      srandom((unsigned long)atoi(optarg));
      break;
    case 't':
      o.t = optarg[0];
      func = do_time;
      break;
    case '?':
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  o.argv = argv;
  o.argc = argc;
  
  func(&o);

  return 0;
}
